import unittest
from abtoolkit.continuous.simulation import StatTestsSimulation
from abtoolkit.utils import generate_data


class TestStatTestsSimulation(unittest.TestCase):
    tests = ["ttest", "diff_ttest", "regression_test", "cuped_ttest",
             "did_regression_test", "additional_vars_regression_test"]

    def test_success(self):
        variable = generate_data(100, distribution_type="cont")
        previous_value = generate_data(100, distribution_type="cont", index=variable.index).rename("prev")

        experiments_num = 10

        sim = StatTestsSimulation(
            variable,
            stattests_list=self.tests,
            experiments_num=experiments_num,
            alternative="two-sided",
            sample_size=50,
            mde=10,
            alpha_level=0.05,
            power=0.8,

            previous_values=previous_value,
            cuped_covariant=previous_value,
            additional_vars=[previous_value],
        )
        info = sim.run()

        self.assertTrue(len(self.tests) == len(info.keys()),
                        "Number of tests doesn't match with number of result dicts")
        for test in self.tests:
            self.assertTrue(test in info.keys(), f"Test '{test}' is not in result dict")
            test_info = info[test]
            self.assertTrue("alpha" in test_info, f"Alpha value not in result dict")
            self.assertTrue(0 <= test_info["alpha"] <= 1, f"Alpha value has wrong value: {test_info['alpha']}")

            self.assertTrue("alpha_ci" in test_info, f"Alpha CI value not in result dict")
            self.assertTrue(len(test_info["alpha_ci"]) == 2,
                            f"Wrong confidence interval for alpha: {test_info['alpha_ci']}")

            self.assertTrue("power" in test_info, f"Power value not in result dict")
            self.assertTrue(0 <= test_info["power"] <= 1, f"Power value has wrong value: {test_info['power']}")

            self.assertTrue("power_ci" in test_info, f"Power CI value not in result dict")
            self.assertTrue(len(test_info["power_ci"]) == 2,
                            f"Wrong confidence interval for power: {test_info['power_ci']}")

            self.assertTrue("aa_pvalues" in test_info, f"AA p-values not in result dict")
            self.assertTrue(len(test_info["aa_pvalues"]) == experiments_num,
                            f"Number of p-values in AA test doesn't match with number of experiments")

            self.assertTrue("ab_pvalues" in test_info, f"AB p-values not in result dict")
            self.assertTrue(len(test_info["ab_pvalues"]) == experiments_num,
                            f"Number of p-values in AB test doesn't match with number of experiments")



import unittest
import numpy as np
from abtoolkit.utils import check_clt


class TestUtils(unittest.TestCase):
    def test_check_central_limit_theorem_normal(self):
        var = np.random.normal(0, 2, size=10000)
        p_value = check_clt(var, do_plot_distribution=False, sample_size=10)
        self.assertTrue(p_value >= 0, f"p-value: {p_value}")

    def test_check_central_limit_theorem_continuous(self):
        var = np.random.chisquare(df=2, size=10000)
        p_value = check_clt(var, do_plot_distribution=False, sample_size=10000)
        self.assertTrue(p_value >= 0, f"p-value: {p_value}")

    def test_check_central_limit_theorem_discrete(self):
        var = np.random.choice([0, 1], p=[0.93, 0.07], size=10000)
        p_value = check_clt(var, do_plot_distribution=False, sample_size=50000)
        self.assertTrue(p_value >= 0, f"p-value: {p_value}")

    def test_check_central_limit_theorem_normal_median(self):
        var = np.random.normal(0, 2, size=10000)
        p_value = check_clt(var, do_plot_distribution=False, sample_size=10, metric_f=np.median)
        self.assertTrue(p_value >= 0, f"p-value: {p_value}")

    def test_check_central_limit_theorem_normal_percentile(self):
        var = np.random.normal(0, 2, size=10000)
        p_value = check_clt(var, do_plot_distribution=False, metric_f=np.percentile, q=0.9)
        self.assertTrue(p_value >= 0, f"p-value: {p_value}")



import unittest
from abtoolkit.discrete.utils import estimate_sample_size_by_mde
from abtoolkit.discrete.utils import estimate_mde_by_sample_size
from abtoolkit.discrete.utils import estimate_ci_binomial


class TestSampleSizeEstimation(unittest.TestCase):
    def test_calculate_sample_size_by_mde(self):
        sample_size = estimate_sample_size_by_mde(0.07, 0.05, 0.8, 0.03, alternative="two-sided")
        self.assertTrue(isinstance(sample_size, int), "Estimated sample size has wrong type")
        self.assertTrue(sample_size > 0, f"Estimated sample size has negative value: {sample_size}")

    def test_calculate_mde_by_sample_size(self):
        mde = estimate_mde_by_sample_size(0.07, 0.05, 0.8, 1136, alternative="two-sided")
        self.assertTrue(isinstance(mde, float), "Estimated MDE has wrong type")
        self.assertTrue(mde > 0, f"Estimated MDE is negative: {mde}")

    def test_estimate_ci_binomial(self):
        ci = estimate_ci_binomial(0.07, 1136, 0.5)
        self.assertTrue(len(ci) == 2, f"Got wrong confidence interval: {ci}")
        self.assertTrue(ci[0] > 0, f"Estimated ci has negative value: {ci[0]}")
        self.assertTrue(ci[1] > 0, f"Estimated ci has negative value: {ci[1]}")



from abtoolkit.utils import check_clt
import numpy as np


if __name__ == '__main__':
    print(f"Checking Central Limit Theorem for Chi-Squared distributed variable")
    chi_variable = np.random.chisquare(df=1, size=10000)
    for sample_size in [10, 100, 1000, 5000]:
        pvalue = check_clt(chi_variable, sample_size=sample_size)
        print(f"\nSample_size = {sample_size}", end='\t')
        if pvalue < 0.05:
            print("\033[91m" + f"Fail, p-value={pvalue}" + "\033[0m")
        else:
            print("\033[92m" + f"Pass, p-value={pvalue}" + "\033[0m")

    print(f"Checking Central Limit Theorem for Normal distributed variable")
    normal_variable = np.random.normal(loc=0, scale=1, size=10000)
    for sample_size in [2, 10, 100]:
        pvalue = check_clt(normal_variable, sample_size=sample_size)
        print(f"\nSample_size = {sample_size}", end='\t')
        if pvalue < 0.05:
            print("\033[91m" + f"Fail, p-value={pvalue}" + "\033[0m")
        else:
            print("\033[92m" + f"Pass, p-value={pvalue}" + "\033[0m")



"""
Stat tests for continuous variables
"""

from typing import List, Literal

import linearmodels as lm
import numpy as np
import pandas as pd
from scipy import special


def _corrected_regression_p_value(
    value: float,
    p_value: float,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Correct regression p-value according to alternative hypothesis.
    If value < 0 then alternative hypothesis = "less" should fail and otherwise,
    If value > 0 then alternative hypothesis = "greater" should fail.
    :param value: value of weight
    :param p_value: p-value of significant of weight
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: corrected p_value
    """

    if (alternative == "less") and (value < 0):
        p_value = 1
    elif (alternative == "greater") and (value > 0):
        p_value = 1

    return p_value


def ttest(
    control: pd.Series,
    test: pd.Series,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Simple t-test
    :param control: pd.Series for control sample
    :param test: pd.Series for test sample
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: p-value
    """

    n1, n2 = len(control), len(test)  # Samples num
    v1, v2 = control.var(), test.var()  # Variance
    m1, m2 = control.mean(), test.mean()  # Mean

    df = n1 + n2 - 2
    if df < 1:
        raise ValueError(f"df = {df}, too few samples in dataset")

    se = ((n1 - 1) * v1 + (n2 - 1) * v2) / df
    t = (m1 - m2) / np.sqrt(se * (1 / n1 + 1 / n2))

    if alternative == "less":
        p_value = special.stdtr(df, t)
    elif alternative == "greater":
        p_value = special.stdtr(df, -t)
    elif alternative == "two-sided":
        p_value = special.stdtr(df, -np.abs(t)) * 2
    else:
        raise ValueError("alternative must be 'less', 'greater' or 'two-sided'")

    return p_value


def difference_ttest(
    control: pd.Series,
    control_pre: pd.Series,
    test: pd.Series,
    test_pre: pd.Series,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Estimation treatment effect using ttest and CUPED to increase test's power
    :param control: pd.Series, control sample
    :param control_pre: pd.Series, control previous period value
    :param test: pd.Series, test sample
    :param test_pre: pd.Series, test previous period value
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: p-value
    """
    control = control - control_pre
    test = test - test_pre

    return ttest(control, test, alternative)


def cuped_ttest(
    control: pd.Series,
    control_covariant: pd.Series,
    test: pd.Series,
    test_covariant: pd.Series,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Estimation treatment effect using ttest and CUPED to increase test's power
    :param control: pd.Series, control sample
    :param control_covariant: pd.Series, control sample covariant
    :param test: pd.Series, test sample
    :param test_covariant: pd.Series, test sample covariant
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: p-value
    """

    full_value = pd.concat(
        [
            control.rename("value"),
            test.rename("value"),
        ],
        axis=0,
    )

    full_covariant = pd.concat(
        [
            control_covariant.rename("covariant"),
            test_covariant.rename("covariant"),
        ],
        axis=0,
    )

    cov = np.cov(full_covariant, full_value)[0, 1]
    var = full_covariant.var()
    theta = cov / var

    cuped_test = test - theta * test_covariant
    cuped_control = control - theta * control_covariant

    return ttest(cuped_control, cuped_test, alternative)


def regression_test(
    control: pd.Series,
    test: pd.Series,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Treatment effect estimation using linear regression
    :param control: pd.Series with index [entity, dt], where dt could be int of datetime. Control sample
    :param test: pd.Series with index [entity, dt], where dt could be int of datetime. Test sample
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: p-value
    """
    df = pd.concat(
        [
            control.rename("value").to_frame().assign(treated=0),
            test.rename("value").to_frame().assign(treated=1),
        ],
        axis=0,
    )
    df["bias"] = 1

    if not isinstance(df.index, pd.MultiIndex):
        df["index1"] = 0
        df["index2"] = 1
        df = df.set_index(["index1", "index2"])

    mod = lm.PanelOLS.from_formula("value ~ bias + treated", data=df)
    result = mod.fit()
    return _corrected_regression_p_value(result.params["treated"], result.pvalues["treated"], alternative)


def did_regression_test(
    control: pd.Series,
    control_pre: pd.Series,
    test: pd.Series,
    test_pre: pd.Series,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Difference-in-Difference treatment effect estimation using linear regression.
    Calculates difference between current and last values in test and control groups and then
    calculates difference between differences to increase test power
    :param control_pre: pd.Series with index [entity, dt], where dt could be int of datetime.
    Control sample before treatment
    :param control: pd.Series with index [entity, dt], where dt could be int of datetime.
    Control sample after treatment
    :param test_pre: pd.Series with index [entity, dt], where dt could be int of datetime. Test sample before treatment
    :param test: pd.Series with index [entity, dt], where dt could be int of datetime. Test sample after treatment
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: p-value
    """
    df = pd.concat(
        [
            control_pre.rename("value").to_frame().assign(treated=0).assign(after=0),
            control.rename("value").to_frame().assign(treated=0).assign(after=1),
            test_pre.rename("value").to_frame().assign(treated=1).assign(after=0),
            test.rename("value").to_frame().assign(treated=1).assign(after=1),
        ],
        axis=0,
    )

    df["bias"] = 1

    if not isinstance(df.index, pd.MultiIndex):
        df["index1"] = 0
        df["index2"] = 1
        df = df.set_index(["index1", "index2"])

    mod = lm.PanelOLS.from_formula("value ~ bias + after + treated + treated*after", data=df)
    result = mod.fit()
    return _corrected_regression_p_value(result.params["treated:after"], result.pvalues["treated:after"], alternative)


def additional_vars_regression_test(
    control: pd.Series,
    control_additional_vars: List[pd.Series],
    test: pd.Series,
    test_additional_vars: List[pd.Series],
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Treatment effect estimation using additional variables in linear regression. Additional
    variables should reduce deviation of target variable and increase test power
    :param control: pd.Series with index [entity, dt], where dt could be int of datetime.
    Control sample
    :param control_additional_vars: List of pd.Series with index [entity, dt], where dt could be int of datetime.
    Additional variables which can describe some deviation of tested variable
    :param test: pd.Series with index [entity, dt], where dt could be int of datetime.
    Test sample
    :param test_additional_vars: List of pd.Series with index [entity, dt], where dt could be int of datetime.
    Additional variables which can describe some deviation of tested variable
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: p-value
    """

    assert len(test_additional_vars) > 0, "No additional vars for 'additional_vars_regression_test' test given"

    additional_vars_names_test = [v.name for v in test_additional_vars]
    additional_vars_names_control = [v.name for v in control_additional_vars]
    assert set(additional_vars_names_test) == set(additional_vars_names_control), (
        f"Lists of control and test additional vars should the same. "
        f"Got {set(additional_vars_names_test)} vars for test "
        f"and {set(additional_vars_names_control)} vars for control"
    )

    control_df = pd.concat([control.rename("value").to_frame()] + control_additional_vars, axis=1)
    test_df = pd.concat([test.rename("value").to_frame()] + test_additional_vars, axis=1)

    control_df.index = test_df.index
    df = pd.concat(
        [
            control_df.assign(treated=0),
            test_df.assign(treated=1),
        ],
        axis=0,
    )

    df["bias"] = 1

    if not isinstance(df.index, pd.MultiIndex):
        df["index1"] = 0
        df["index2"] = 1
        df = df.set_index(["index1", "index2"])

    additional_vars_formula = " + ".join(map(str, additional_vars_names_test))

    formula = f"value ~ bias + treated + {additional_vars_formula}"
    mod = lm.PanelOLS.from_formula(formula, data=df)
    result = mod.fit()
    return _corrected_regression_p_value(result.params["treated"], result.pvalues["treated"], alternative)



"""
Stat tests for discrete variable analysis
"""

from typing import Literal

import numpy as np
from scipy import stats


def conversion_ztest(
    control_count: int,
    control_objects_num: int,
    test_count: int,
    test_objects_num: int,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Conversion z-test
    :param control_count: number of positive samples in control group
    :param control_objects_num: number of all samples in control group
    :param test_count: number of positive samples in test group
    :param test_objects_num: number of all samples in test group
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : conversions are equal;
    * 'less': the conversion of the control sample is less than the mean of the test sample;
    * 'greater': the conversion of the control sample is greater than the mean of the test sample;
    :return: p-value
    """

    diff = control_count / control_objects_num - test_count / test_objects_num
    p_pooled = (control_count + test_count) / (control_objects_num + test_objects_num)
    std_diff = np.sqrt(p_pooled * (1 - p_pooled) * (1 / control_objects_num + 1 / test_objects_num))
    z_stat = diff / std_diff

    if alternative == "less":
        p_value = stats.norm.cdf(z_stat)
    elif alternative == "greater":
        p_value = stats.norm.sf(z_stat)
    elif alternative == "two-sided":
        p_value = stats.norm.sf(np.abs(z_stat)) * 2
    else:
        raise ValueError("alternative must be 'less', 'greater' or 'two-sided'")

    return p_value



"""
Simulates AA and AB tests to estimate test power and alpha
"""

from typing import List, Literal

import numpy as np

from abtoolkit.discrete.stattests import conversion_ztest
from abtoolkit.utils import BaseSimulationClass


class StatTestsSimulation(BaseSimulationClass):
    """
    Class simulates AA and AB tests to estimate test power, builds p-value distribution plot
    """

    def __init__(
        self,
        count: int,
        objects_num: int,
        alternative: Literal["less", "greater", "two-sided"],
        stattests_list: List[str],
        sample_size: int,
        experiments_num: int,
        mde: float,
        alpha_level: float = 0.05,
        power: float = 0.8,
    ):
        """
        Simulates AA and AB tests for given stat-tests. Prints result (alpha and power) for each test
        and builds plot for p-value distributions.

        :param count: number of positive samples in dataset
        :param objects_num: number of all samples in dataset
        :param stattests_list: list of stat-tests for estimation
        :param sample_size: number of examples to sample from variables in each iteration
        :param experiments_num: number of experiments to perform for each stat-test
        :param mde: minimal detectable effect, used to perform AB test (add to test variable)
        :param alpha_level: test alpha-level
        :param power: test power
        """
        super().__init__(
            sample_size=sample_size,
            alternative=alternative,
            mde=mde,
            power=power,
            alpha_level=alpha_level,
            stattests_list=stattests_list,
            experiments_num=experiments_num,
        )

        self.count = count
        self.objects_num = objects_num
        self.p = self.count / self.objects_num
        self.stattests_func_map = {
            "conversion_ztest": self.simulate_conversion_ztest,
        }

    def simulate_conversion_ztest(self, mde: float) -> float:
        """
        Simulate ttest
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """

        control_count = np.random.binomial(n=self.sample_size, p=self.p)
        test_count = np.random.binomial(n=self.sample_size, p=self.p + mde)

        return conversion_ztest(control_count, self.sample_size, test_count, self.sample_size, self.alternative)



import unittest

import numpy as np
from abtoolkit.discrete.stattests import conversion_ztest
from abtoolkit.utils import generate_data


class TestStatTests(unittest.TestCase):
    def test_ttest(self):
        test_sr = generate_data(100, distribution_type="disc")
        control_sr = generate_data(100, distribution_type="disc")
        p_value = conversion_ztest(
            control_sr.sum(),
            len(control_sr),
            test_sr.sum(),
            len(test_sr),
            "less")
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")



import unittest

from abtoolkit.continuous.stattests import regression_test
from abtoolkit.continuous.stattests import did_regression_test
from abtoolkit.continuous.stattests import additional_vars_regression_test
from abtoolkit.continuous.stattests import cuped_ttest
from abtoolkit.continuous.stattests import difference_ttest
from abtoolkit.continuous.stattests import ttest
from abtoolkit.utils import generate_data


class TestStatTests(unittest.TestCase):
    def test_ttest(self):
        test_sr = generate_data(100, distribution_type="cont")
        control_sr = generate_data(100, distribution_type="cont")
        p_value = ttest(control_sr, test_sr, "two-sided")
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_regression_test(self):
        test_sr = generate_data(100, distribution_type="cont")
        control_sr = generate_data(100, distribution_type="cont")
        p_value = regression_test(control_sr, test_sr, "two-sided")
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_diff_ttest(self):
        test_sr = generate_data(100, distribution_type="cont")
        test_pre_sr = generate_data(100, distribution_type="cont", index=test_sr.index).rename("var_1")
        control_sr = generate_data(100, distribution_type="cont")
        control_pre_sr = generate_data(100, distribution_type="cont", index=control_sr.index).rename("var_1")
        p_value = difference_ttest(control_pre_sr, control_sr, test_pre_sr, test_sr, "two-sided")
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_did_regression_test(self):
        test_pre_sr = generate_data(100, distribution_type="cont")
        test_post_sr = generate_data(100, distribution_type="cont")
        control_pre_sr = generate_data(100, distribution_type="cont")
        control_post_sr = generate_data(100, distribution_type="cont")
        p_value = did_regression_test(control_pre_sr, control_post_sr, test_pre_sr, test_post_sr, 
                                      "two-sided")
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_additional_vars_reg_one_var_test(self):
        test_sr = generate_data(100, distribution_type="cont")
        test_additional_var = generate_data(100, distribution_type="cont", index=test_sr.index).rename("var_1")
        control_sr = generate_data(100, distribution_type="cont")
        control_additional_var = generate_data(100, distribution_type="cont", index=control_sr.index).rename("var_1")
        p_value = additional_vars_regression_test(
            control_sr,
            [control_additional_var],
            test_sr,
            [test_additional_var],
            "two-sided"
        )
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_additional_vars_reg_one_var_test_fail(self):
        test_sr = generate_data(100, distribution_type="cont")
        test_additional_var = generate_data(100, distribution_type="cont", index=test_sr.index).rename("var_1")
        control_sr = generate_data(100, distribution_type="cont")
        control_additional_var = generate_data(100, distribution_type="cont", index=control_sr.index).rename("var_2")
        self.assertRaises(AssertionError, additional_vars_regression_test,
                          control_sr, [control_additional_var],
                          test_sr, [test_additional_var], "two-sided")

    def test_additional_vars_reg_multiple_vars_test(self):
        test_sr = generate_data(100, distribution_type="cont")
        test_additional_var1 = generate_data(100, distribution_type="cont", index=test_sr.index).rename("var_1")
        test_additional_var2 = generate_data(100, distribution_type="cont", index=test_sr.index).rename("var_1")
        control_sr = generate_data(100, distribution_type="cont")
        control_additional_var1 = generate_data(100, distribution_type="cont", index=control_sr.index).rename("var_1")
        control_additional_var2 = generate_data(100, distribution_type="cont", index=control_sr.index).rename("var_1")
        p_value = additional_vars_regression_test(
            control_sr,
            [control_additional_var1, control_additional_var2],
            test_sr,
            [test_additional_var1, test_additional_var2],
            "two-sided"
        )
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_additional_vars_reg_no_vars_test(self):
        test_sr = generate_data(100, distribution_type="cont")
        control_sr = generate_data(100, distribution_type="cont")
        self.assertRaises(AssertionError, additional_vars_regression_test, control_sr, [], test_sr, [], "two-sided")

    def test_cuped_ttest(self):
        test_sr = generate_data(100, distribution_type="cont")
        test_covariant_sr = generate_data(100, distribution_type="cont", index=test_sr.index).rename("var_1")
        control_sr = generate_data(100, distribution_type="cont")
        control_covariant_sr = generate_data(100, distribution_type="cont", index=control_sr.index).rename("var_1")
        p_value = cuped_ttest(
            control_sr,
            control_covariant_sr,
            test_sr,
            test_covariant_sr,
            "two-sided",
        )
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")



"""
Utils functions for discrete variable analysis
"""

from typing import Literal

import numpy as np
from scipy import stats


def estimate_sample_size_by_mde(
    p: float,
    alpha: float,
    power: float,
    mde: float,
    alternative: Literal["less", "greater", "two-sided"],
) -> int:
    """
    Estimate sample size need for significant test using p, alpha, power, mde
    :param p: probability of positive sample
    :param alpha: level of significance of A/B test
    :param power: probability of observing a statistically significant result at level alpha
    if a true effect of a certain magnitude is present
    :param mde: minimal detectable effect, the difference in results to detect
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: sample size needed for each group
    """

    # Todo: When alternative = "less" or "greater" we get undersampling (power less then 0.8).
    if alternative == "two-sided":
        pass

    alpha = alpha / 2

    z = stats.norm.ppf(q=1 - alpha) + stats.norm.ppf(q=power)
    size = 2 * p * (1 - p) * (z / mde) ** 2
    return round(size) + 1


def estimate_mde_by_sample_size(
    p: float,
    alpha: float,
    power: float,
    sample_size: int,
    alternative: Literal["less", "greater", "two-sided"],
) -> float:
    """
    Estimate MDE by need for significant test using p, alpha, power, sample_size
    :param p: probability of positive sample
    :param alpha: level of significance of A/B test
    :param power: probability of observing a statistically significant result at level alpha
    if a true effect of a certain magnitude is present
    :param sample_size: number of users in each group (test and control)
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: minimum detectable effect
    """

    # Todo: When alternative = "less" or "greater" we get undersampling (power less then 0.8).
    if alternative == "two-sided":
        pass

    alpha = alpha / 2

    z = stats.norm.ppf(q=1 - alpha) + stats.norm.ppf(q=power)
    # size = 2 * p * (1 - p) * (z / mde) ** 2

    mde = z / np.sqrt(sample_size / (2 * p * (1 - p)))

    return mde


def estimate_ci_binomial(p, sample_size, alpha):
    """
    Confidence interval for Binomial variable
    :param p: probability of positive
    :param sample_size: number of samples
    :param alpha: alpha level
    :return: low confidence interval value, high confidence interval value
    """
    t = stats.norm.ppf(1 - alpha / 2, loc=0, scale=1)
    std_n = np.sqrt(p * (1 - p) / sample_size)
    return p - t * std_n, p + t * std_n



"""
Simulates AA and AB tests to estimate test power and alpha
"""

from typing import List, Literal

import numpy as np
import pandas as pd

from abtoolkit.continuous.stattests import additional_vars_regression_test
from abtoolkit.continuous.stattests import cuped_ttest
from abtoolkit.continuous.stattests import did_regression_test
from abtoolkit.continuous.stattests import regression_test
from abtoolkit.continuous.stattests import ttest
from abtoolkit.utils import BaseSimulationClass


class StatTestsSimulation(BaseSimulationClass):
    """
    Class simulates AA and AB tests to estimate test power, builds p-value distribution plot
    """

    def __init__(
        self,
        variable: pd.Series,
        alternative: Literal["less", "greater", "two-sided"],
        stattests_list: List[str],
        sample_size: int,
        experiments_num: int,
        mde: float,
        alpha_level: float = 0.05,
        power: float = 0.8,
        previous_values: pd.Series = None,
        cuped_covariant: pd.Series = None,
        additional_vars: List[pd.Series] = None,
    ):
        """
        Simulates AA and AB tests for given stat-tests. Prints result (alpha and power) for each test
        and builds plot for p-value distributions.

        :param variable: variable for simulation
        :param stattests_list: list of stat-tests for estimation
        :param sample_size: number of examples to sample from variables in each iteration
        :param experiments_num: number of experiments to perform for each stat-test
        :param mde: minimal detectable effect, used to perform AB test (add to test variable)
        :param alpha_level: test alpha-level
        :param power: test power
        :param previous_values: previous values of variable used to reduce variance and speedup
        test in difference-in-difference test
        :param cuped_covariant: covariant for variable used to reduce variance and speedup test
        in cuped test
        :param additional_vars: list of additional variables used to
        reduce variance of main variable and speedup test in 'regression_with_additional_variables' test
        """
        super().__init__(
            sample_size=sample_size,
            alternative=alternative,
            mde=mde,
            power=power,
            alpha_level=alpha_level,
            stattests_list=stattests_list,
            experiments_num=experiments_num,
        )

        self.variable = variable
        self.stattests_func_map = {
            "ttest": self.simulate_ttest,
            "diff_ttest": self.simulate_difference_ttest,
            "cuped_ttest": self.simulate_cuped,
            "regression_test": self.simulate_reg,
            "did_regression_test": self.simulate_reg_did,
            "additional_vars_regression_test": self.simulate_reg_add,
        }

        # Optional
        self.previous_values = previous_values
        self.cuped_covariant = cuped_covariant
        self.additional_vars = additional_vars

    def simulate_ttest(self, mde: float) -> float:
        """
        Simulate ttest
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """
        control_sample = self.variable.sample(self.sample_size, replace=True)
        test_sample = self.variable.sample(self.sample_size, replace=True)
        test_sample += mde

        return ttest(control_sample, test_sample, self.alternative)

    def simulate_difference_ttest(self, mde: float) -> float:
        """
        Simulate ttest for difference between actual variable value and previous period variable value
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """
        control_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]
        test_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]

        control_sample = self.variable.loc[control_index_sample]
        control_pre_sample = self.previous_values.loc[control_index_sample]
        test_sample = self.variable.loc[test_index_sample]
        test_pre_sample = self.previous_values.loc[test_index_sample]
        test_sample += mde

        return cuped_ttest(control_sample, control_pre_sample, test_sample, test_pre_sample, self.alternative)

    def simulate_cuped(self, mde: float) -> float:
        """
        Simulate CUPED ttest
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """
        control_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]
        test_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]

        control_sample = self.variable.loc[control_index_sample]
        control_covariant_sample = self.cuped_covariant.loc[control_index_sample]
        test_sample = self.variable.loc[test_index_sample]
        test_covariant_sample = self.cuped_covariant.loc[test_index_sample]
        test_sample += mde

        return cuped_ttest(
            control_sample, control_covariant_sample, test_sample, test_covariant_sample, self.alternative
        )

    def simulate_reg(self, mde: float) -> float:
        """
        Simulate test using regression
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """
        control_sample = self.variable.sample(self.sample_size, replace=True)
        test_sample = self.variable.sample(self.sample_size, replace=True)
        test_sample += mde

        return regression_test(control_sample, test_sample, self.alternative)

    def simulate_reg_did(self, mde: float) -> float:
        """
        Simulate test using regression with difference-in-difference technique
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """
        control_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]
        test_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]

        control_sample = self.variable.loc[control_index_sample]
        control_previous_sample = self.previous_values.loc[control_index_sample]
        test_sample = self.variable.loc[test_index_sample]
        test_previous_sample = self.previous_values.loc[test_index_sample]
        test_sample += mde

        return did_regression_test(
            control_sample, control_previous_sample, test_sample, test_previous_sample, self.alternative
        )

    def simulate_reg_add(self, mde: float) -> float:
        """
        Simulate test using regression with additional variables
        :param mde: minimal detectable effect, to sum with test variable
        :return: p_value
        """
        control_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]
        test_index_sample = self.variable.index[np.random.randint(0, len(self.variable), size=self.sample_size)]

        control_sample = self.variable.loc[control_index_sample]
        control_add_samples = [a.loc[control_index_sample] for a in self.additional_vars]
        test_sample = self.variable.loc[test_index_sample]
        test_add_samples = [a.loc[test_index_sample] for a in self.additional_vars]
        test_sample += mde

        return additional_vars_regression_test(
            control_sample, control_add_samples, test_sample, test_add_samples, self.alternative
        )
